"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.OtpController = void 0;
const tslib_1 = require("tslib");
// Copyright (c) 2023 Sourcefuse Technologies
//
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT
const context_1 = require("@loopback/context");
const repository_1 = require("@loopback/repository");
const rest_1 = require("@loopback/rest");
const core_1 = require("@sourceloop/core");
const jwt = tslib_1.__importStar(require("jsonwebtoken"));
const loopback4_authentication_1 = require("loopback4-authentication");
const loopback4_authorization_1 = require("loopback4-authorization");
const otplib_1 = require("otplib");
const qrcode_1 = tslib_1.__importDefault(require("qrcode"));
const keys_1 = require("../../../keys");
const providers_1 = require("../../../providers");
const repositories_1 = require("../../../repositories");
const models_1 = require("../models");
let OtpController = class OtpController {
    constructor(authClientRepository, userRepo, otpCacheRepo, userCredsRepository, logger, checkMfa, jwtVerifier, mfaConfig, otpConfig) {
        this.authClientRepository = authClientRepository;
        this.userRepo = userRepo;
        this.otpCacheRepo = otpCacheRepo;
        this.userCredsRepository = userCredsRepository;
        this.logger = logger;
        this.checkMfa = checkMfa;
        this.jwtVerifier = jwtVerifier;
        this.mfaConfig = mfaConfig;
        this.otpConfig = otpConfig;
    }
    // OTP
    async sendOtp(req) {
        // This is intentional
    }
    async verifyOtp(req, user, codeWriter) {
        let otpCache;
        let clientId;
        let userId;
        let clientSecret;
        const isMfaEnabled = await this.checkMfa(user);
        if (isMfaEnabled) {
            if (this.mfaConfig.secondFactor === "otp" /* STRATEGY.OTP */ &&
                this.otpConfig.method === "OTP" /* OtpMethodType.OTP */) {
                otpCache = await this.otpCacheRepo.get(req.key);
                if (user === null || user === void 0 ? void 0 : user.id) {
                    otpCache.userId = user.id;
                }
                clientId = otpCache.clientId;
                userId = otpCache.userId;
                clientSecret = otpCache.clientSecret;
            }
            else if (this.mfaConfig.secondFactor === "otp" /* STRATEGY.OTP */ &&
                this.otpConfig.method === "Google Authenticator App" /* OtpMethodType.GOOGLE_AUTHENTICATOR */) {
                if (!req.clientId) {
                    throw new Error('Client ID must be provided for Google Authenticator');
                }
                clientId = req.clientId;
                userId = user.id;
                const client = await this.authClientRepository.findOne({
                    where: {
                        clientId,
                    },
                });
                clientSecret = client === null || client === void 0 ? void 0 : client.clientSecret;
            }
            else {
                // This is intentional.
            }
        }
        const codePayload = {
            clientId: clientId,
            userId: userId,
        };
        const token = await codeWriter(jwt.sign(codePayload, clientSecret, {
            expiresIn: 180,
            audience: clientId,
            issuer: process.env.JWT_ISSUER,
            algorithm: 'HS256',
        }));
        return {
            code: token,
        };
    }
    // Google Authenticator
    async checkQr(code, clientId, codeReader) {
        var _a, _b;
        if (!clientId) {
            throw new rest_1.HttpErrors.Unauthorized("Client Invalid" /* AuthErrorKeys.ClientInvalid */);
        }
        if (!code) {
            throw new rest_1.HttpErrors.Unauthorized("Token Invalid" /* AuthErrorKeys.TokenInvalid */);
        }
        const authClient = await this.authClientRepository.findOne({
            where: {
                clientId: clientId,
            },
        });
        if (!authClient) {
            throw new rest_1.HttpErrors.Unauthorized("Client Invalid" /* AuthErrorKeys.ClientInvalid */);
        }
        try {
            const authCode = await codeReader(code);
            const payload = (await this.jwtVerifier(authCode, {
                audience: clientId,
            }));
            const userId = (_a = payload.userId) !== null && _a !== void 0 ? _a : (_b = payload.user) === null || _b === void 0 ? void 0 : _b.id;
            const userCreds = await this.userCredsRepository.findOne({
                where: {
                    userId: userId,
                },
                fields: {
                    secretKey: true,
                },
            });
            if (!userCreds) {
                throw new rest_1.HttpErrors.Unauthorized("UserDoesNotExist" /* AuthenticateErrorKeys.UserDoesNotExist */);
            }
            return {
                isGenerated: Boolean(userCreds.secretKey),
            };
        }
        catch (error) {
            this.logger.error(error);
            if (error.name === 'TokenExpiredError') {
                throw new rest_1.HttpErrors.Unauthorized("Code Expired" /* AuthErrorKeys.CodeExpired */);
            }
            else if (rest_1.HttpErrors.HttpError.prototype.isPrototypeOf(error)) {
                throw error;
            }
            else {
                throw new rest_1.HttpErrors.Unauthorized("Invalid Credentials" /* AuthErrorKeys.InvalidCredentials */);
            }
        }
    }
    async createQr(req, codeReader) {
        var _a, _b, _c, _d;
        const authClient = await this.authClientRepository.findOne({
            where: {
                clientId: req.clientId,
            },
        });
        if (!authClient) {
            throw new rest_1.HttpErrors.Unauthorized("Client Invalid" /* AuthErrorKeys.ClientInvalid */);
        }
        try {
            const code = await codeReader(req.code);
            const payload = (await this.jwtVerifier(code, {
                audience: req.clientId,
            }));
            const userId = (_a = payload.userId) !== null && _a !== void 0 ? _a : (_b = payload.user) === null || _b === void 0 ? void 0 : _b.id;
            const userCreds = await this.userCredsRepository.findOne({
                where: {
                    userId: userId,
                },
                fields: {
                    id: true,
                },
            });
            if (!userCreds) {
                throw new rest_1.HttpErrors.Unauthorized("UserDoesNotExist" /* AuthenticateErrorKeys.UserDoesNotExist */);
            }
            const secretKey = otplib_1.authenticator.generateSecret();
            await this.userCredsRepository.updateById(userCreds.id, {
                secretKey: secretKey,
            });
            const appName = (_c = process.env.APP_NAME) !== null && _c !== void 0 ? _c : 'auth-service';
            let username = (_d = payload.user) === null || _d === void 0 ? void 0 : _d.username;
            if (!username) {
                const user = await this.userRepo.findById(userId);
                username = user.username;
            }
            const otpauth = otplib_1.authenticator.keyuri(username, appName, secretKey);
            const qrCode = await qrcode_1.default.toDataURL(otpauth);
            return {
                qrCode,
            };
        }
        catch (error) {
            this.logger.error(error);
            if (error.name === 'TokenExpiredError') {
                throw new rest_1.HttpErrors.Unauthorized("Code Expired" /* AuthErrorKeys.CodeExpired */);
            }
            else if (rest_1.HttpErrors.HttpError.prototype.isPrototypeOf(error)) {
                throw error;
            }
            else {
                throw new rest_1.HttpErrors.Unauthorized("Invalid Credentials" /* AuthErrorKeys.InvalidCredentials */);
            }
        }
    }
};
exports.OtpController = OtpController;
tslib_1.__decorate([
    (0, loopback4_authentication_1.authenticateClient)("client-password" /* STRATEGY.CLIENT_PASSWORD */),
    (0, loopback4_authentication_1.authenticate)("otp" /* STRATEGY.OTP */),
    (0, loopback4_authorization_1.authorize)({ permissions: ['*'] }),
    (0, rest_1.post)('/auth/send-otp', {
        description: 'Sends OTP',
        responses: {
            [200 /* STATUS_CODE.OK */]: {
                description: 'Sends otp to user',
                content: {
                    [core_1.CONTENT_TYPE.JSON]: Object,
                },
            },
            ...core_1.ErrorCodes,
        },
    }),
    tslib_1.__param(0, (0, rest_1.requestBody)()),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", [models_1.OtpSendRequest]),
    tslib_1.__metadata("design:returntype", Promise)
], OtpController.prototype, "sendOtp", null);
tslib_1.__decorate([
    (0, loopback4_authentication_1.authenticate)("otp" /* STRATEGY.OTP */),
    (0, loopback4_authorization_1.authorize)({ permissions: ['*'] }),
    (0, rest_1.post)('/auth/verify-otp', {
        description: 'Gets you the code that will be used for getting token (webapps)',
        responses: {
            [200 /* STATUS_CODE.OK */]: {
                description: 'Auth Code that you can use to generate access and refresh tokens using the POST /auth/token API',
                content: {
                    [core_1.CONTENT_TYPE.JSON]: Object,
                },
            },
            ...core_1.ErrorCodes,
        },
    }),
    tslib_1.__param(0, (0, rest_1.requestBody)()),
    tslib_1.__param(1, (0, context_1.inject)(loopback4_authentication_1.AuthenticationBindings.CURRENT_USER)),
    tslib_1.__param(2, (0, context_1.inject)(providers_1.AuthCodeBindings.CODEWRITER_PROVIDER)),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", [models_1.OtpLoginRequest,
        models_1.AuthUser, Function]),
    tslib_1.__metadata("design:returntype", Promise)
], OtpController.prototype, "verifyOtp", null);
tslib_1.__decorate([
    (0, loopback4_authorization_1.authorize)({ permissions: ['*'] }),
    (0, rest_1.get)('/auth/check-qr-code', {
        description: 'Returns isGenerated:true if secret_key already exist',
        responses: {
            [200 /* STATUS_CODE.OK */]: {
                description: 'secret_key already exists',
                content: {
                    [core_1.CONTENT_TYPE.JSON]: Object,
                },
            },
            ...core_1.ErrorCodes,
        },
    }),
    tslib_1.__param(0, rest_1.param.header.string('code')),
    tslib_1.__param(1, rest_1.param.header.string('clientId')),
    tslib_1.__param(2, (0, context_1.inject)(providers_1.AuthCodeBindings.CODEREADER_PROVIDER)),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", [String, String, Function]),
    tslib_1.__metadata("design:returntype", Promise)
], OtpController.prototype, "checkQr", null);
tslib_1.__decorate([
    (0, loopback4_authorization_1.authorize)({ permissions: ['*'] }),
    (0, rest_1.post)('/auth/create-qr-code', {
        description: 'Generates a new qrCode for Authenticator App',
        responses: {
            [200 /* STATUS_CODE.OK */]: {
                description: 'qrCode that you can use to generate codes in Authenticator App',
                content: {
                    [core_1.CONTENT_TYPE.JSON]: Object,
                },
            },
            ...core_1.ErrorCodes,
        },
    }),
    tslib_1.__param(0, (0, rest_1.requestBody)()),
    tslib_1.__param(1, (0, context_1.inject)(providers_1.AuthCodeBindings.CODEREADER_PROVIDER)),
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", [models_1.AuthTokenRequest, Function]),
    tslib_1.__metadata("design:returntype", Promise)
], OtpController.prototype, "createQr", null);
exports.OtpController = OtpController = tslib_1.__decorate([
    tslib_1.__param(0, (0, repository_1.repository)(repositories_1.AuthClientRepository)),
    tslib_1.__param(1, (0, repository_1.repository)(repositories_1.UserRepository)),
    tslib_1.__param(2, (0, repository_1.repository)(repositories_1.OtpCacheRepository)),
    tslib_1.__param(3, (0, repository_1.repository)(repositories_1.UserCredentialsRepository)),
    tslib_1.__param(4, (0, context_1.inject)(core_1.LOGGER.LOGGER_INJECT)),
    tslib_1.__param(5, (0, context_1.inject)(providers_1.VerifyBindings.MFA_PROVIDER)),
    tslib_1.__param(6, (0, context_1.inject)(providers_1.AuthCodeBindings.JWT_VERIFIER.key)),
    tslib_1.__param(7, (0, context_1.inject)(keys_1.AuthServiceBindings.MfaConfig, { optional: true })),
    tslib_1.__param(8, (0, context_1.inject)(keys_1.AuthServiceBindings.OtpConfig, { optional: true })),
    tslib_1.__metadata("design:paramtypes", [repositories_1.AuthClientRepository,
        repositories_1.UserRepository,
        repositories_1.OtpCacheRepository,
        repositories_1.UserCredentialsRepository, Object, Function, Function, Object, Object])
], OtpController);
//# sourceMappingURL=otp.controller.js.map